/*
 * QEMU sPAPR PCI host for VFIO
 *
 * Copyright (c) 2011 Alexey Kardashevskiy, IBM Corporation.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <sys/types.h>
#include <dirent.h>

#include "hw/hw.h"
#include "hw/ppc/spapr.h"
#include "hw/pci-host/spapr.h"
#include "hw/misc/vfio.h"
#include "hw/pci/pci_bus.h"
#include "trace.h"
#include "qemu/error-report.h"

/* sPAPR VFIO */
static Property spapr_phb_vfio_properties[] = {
    DEFINE_PROP_INT32("iommu", sPAPRPHBVFIOState, iommugroupid, -1),
    DEFINE_PROP_UINT8("scan", sPAPRPHBVFIOState, scan, 1),
    DEFINE_PROP_UINT8("mf", sPAPRPHBVFIOState, enable_multifunction, 0),
    DEFINE_PROP_UINT8("forceaddr", sPAPRPHBVFIOState, force_addr, 0),
    DEFINE_PROP_END_OF_LIST(),
};

static void spapr_pci_vfio_scan(sPAPRPHBVFIOState *svphb, Error **errp)
{
    PCIHostState *phb = PCI_HOST_BRIDGE(svphb);
    char *iommupath;
    DIR *dirp;
    struct dirent *entry;
    Error *error = NULL;

    if (!svphb->scan) {
        trace_spapr_pci("autoscan disabled for ", svphb->phb.dtbusname);
        return;
    }

    iommupath = g_strdup_printf("/sys/kernel/iommu_groups/%d/devices/",
                                svphb->iommugroupid);
    if (!iommupath) {
        return;
    }

    dirp = opendir(iommupath);
    if (!dirp) {
        error_report("spapr-vfio: vfio scan failed on opendir: %m");
        g_free(iommupath);
        return;
    }

    while ((entry = readdir(dirp)) != NULL) {
        Error *err = NULL;
        char *tmp;
        FILE *deviceclassfile;
        unsigned deviceclass = 0, domainid, busid, devid, fnid;
        char addr[32];
        DeviceState *dev;

        if (sscanf(entry->d_name, "%X:%X:%X.%x",
                   &domainid, &busid, &devid, &fnid) != 4) {
            continue;
        }

        tmp = g_strdup_printf("%s%s/class", iommupath, entry->d_name);
        trace_spapr_pci("Reading device class from ", tmp);

        deviceclassfile = fopen(tmp, "r");
        if (deviceclassfile) {
            int ret = fscanf(deviceclassfile, "%x", &deviceclass);
            fclose(deviceclassfile);
            if (ret != 1) {
                continue;
            }
        }
        g_free(tmp);

        if (!deviceclass) {
            continue;
        }
        if ((deviceclass >> 16) == (PCI_CLASS_BRIDGE_OTHER >> 8)) {
            /* Skip bridges */
            continue;
        }
        trace_spapr_pci("Creating device from ", entry->d_name);

        dev = qdev_create(&phb->bus->qbus, "vfio-pci");
        if (!dev) {
            trace_spapr_pci("failed to create vfio-pci", entry->d_name);
            continue;
        }
        qdev_prop_parse(dev, "host", entry->d_name, &err);
        if (err != NULL) {
            continue;
        }
        if (svphb->force_addr) {
            snprintf(addr, sizeof(addr), "%x.%x", devid, fnid);
            err = NULL;
            qdev_prop_parse(dev, "addr", addr, &err);
            if (err != NULL) {
                continue;
            }
        }
        if (svphb->enable_multifunction) {
            qdev_prop_set_bit(dev, "multifunction", 1);
        }
        object_property_set_bool(OBJECT(dev), true, "realized", &error);
        if (error) {
            error_propagate(errp, error);
            break;
        }
    }
    closedir(dirp);
    g_free(iommupath);
}

static void spapr_phb_vfio_finish_realize(sPAPRPHBState *sphb, Error **errp)
{
    sPAPRPHBVFIOState *svphb = SPAPR_PCI_VFIO_HOST_BRIDGE(sphb);
    struct vfio_iommu_spapr_tce_info info = { .argsz = sizeof(info) };
    int ret;
    Error *error = NULL;

    if (svphb->iommugroupid == -1) {
        error_setg(errp, "Wrong IOMMU group ID %d", svphb->iommugroupid);
        return;
    }

    svphb->phb.tcet = spapr_vfio_new_table(DEVICE(sphb), svphb->phb.dma_liobn);

    if (!svphb->phb.tcet) {
        error_setg(errp, "spapr-vfio: failed to create VFIO TCE table");
        return;
    }

    address_space_init(&sphb->iommu_as, spapr_tce_get_iommu(sphb->tcet),
                       sphb->dtbusname);

    ret = vfio_container_spapr_get_info(&svphb->phb.iommu_as,
                                        sphb->dma_liobn, svphb->iommugroupid,
                                        &info);
    if (ret) {
        error_setg_errno(errp, -ret,
                         "spapr-vfio: get info from container failed");
        return;
    }

    svphb->phb.dma_window_start = info.dma32_window_start;
    svphb->phb.dma_window_size = info.dma32_window_size;

    spapr_pci_vfio_scan(svphb, &error);
    if (error) {
        error_propagate(errp, error);
    }
}

static void spapr_phb_vfio_reset(DeviceState *qdev)
{
    /* Do nothing */
}

static void spapr_phb_vfio_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);
    sPAPRPHBClass *spc = SPAPR_PCI_HOST_BRIDGE_CLASS(klass);

    dc->props = spapr_phb_vfio_properties;
    dc->reset = spapr_phb_vfio_reset;
    spc->finish_realize = spapr_phb_vfio_finish_realize;
}

static const TypeInfo spapr_phb_vfio_info = {
    .name          = TYPE_SPAPR_PCI_VFIO_HOST_BRIDGE,
    .parent        = TYPE_SPAPR_PCI_HOST_BRIDGE,
    .instance_size = sizeof(sPAPRPHBVFIOState),
    .class_init    = spapr_phb_vfio_class_init,
    .class_size    = sizeof(sPAPRPHBClass),
};

static void spapr_pci_vfio_register_types(void)
{
    type_register_static(&spapr_phb_vfio_info);
}

type_init(spapr_pci_vfio_register_types)
